import React, { ChangeEvent, FormEvent, MouseEvent, PureComponent } from 'react';
import { DateTime, SelectableValue } from '@grafana/data';
import { FormField, FormLabel } from '@grafana/ui';
import { DashboardPicker } from 'app/core/components/Select/DashboardPicker';
import { ReportScheduling } from './ReportScheduling';
import { createReport, updateReport } from './state/actions';
import { Report, ReportDTO, SchedulingFrequency, SchedulingOptions } from '../types';

type OnChangeEditbleStateKey = keyof Pick<State, 'name' | 'replyTo' | 'recipients' | 'message'>;

export interface Props {
  heading: string;
  report?: Report;

  createOrUpdate: typeof createReport | typeof updateReport;
}

export interface State {
  id: number;
  name: string;
  recipients: string;
  replyTo: string;
  message: string;
  dashboardId: number;
  dashboardName: string;
  schedule: SchedulingOptions;
}

function isEmail(email: string) {
  if (!email) {
    return true;
  }

  return email.includes('@');
}

export class ReportPage extends PureComponent<Props, State> {
  constructor(props: Props) {
    super(props);

    this.state = {
      id: props.report.id,
      name: props.report.name,
      recipients: props.report.recipients,
      replyTo: props.report.replyTo,
      message: props.report.message,
      dashboardId: props.report.dashboardId,
      dashboardName: props.report.dashboardName,
      schedule: props.report.schedule,
    };
  }

  onChange = (stateKey: OnChangeEditbleStateKey, value: string | number) => {
    this.setState({
      ...this.state,
      [stateKey]: value,
    });
  };

  onModeChange = (frequency: SchedulingFrequency) => {
    this.setState(prevState => ({
      ...prevState,
      schedule: {
        ...prevState.schedule,
        frequency: frequency,
      },
    }));
  };

  onDayOfWeekChange = (dayOfWeek: SelectableValue<string>) => {
    this.setState(prevState => ({
      ...prevState,
      schedule: {
        ...prevState.schedule,
        day: dayOfWeek.value,
      },
    }));
  };

  onTimeOfDayChange = (timeOfDay: DateTime) => {
    this.setState(prevState => ({
      ...prevState,
      schedule: {
        ...prevState.schedule,
        hour: timeOfDay.hour(),
        minute: timeOfDay.minute(),
      },
    }));
  };

  onDashboardChange = (dashboard: SelectableValue<number>) => {
    if (dashboard) {
      this.setState({
        dashboardId: dashboard.id,
        dashboardName: dashboard.label,
      });
    }
  };

  submitForm = (event: FormEvent<HTMLFormElement> | MouseEvent<HTMLButtonElement>) => {
    event.preventDefault();

    if (this.isFormValid()) {
      const { name, recipients, dashboardId, replyTo, message, schedule } = this.state;
      const { createOrUpdate, report } = this.props;

      const reportDto: ReportDTO = {
        id: report ? report.id : undefined,
        name,
        recipients,
        dashboardId,
        replyTo,
        message,
        schedule,
      };

      createOrUpdate(reportDto);
    }
  };

  validateRecipients = () => {
    const { recipients } = this.state;

    if (!recipients) {
      return true;
    }
    const splitRecipients = recipients.split(';');

    return splitRecipients.every(email => isEmail(email));
  };

  isFormValid() {
    const { name, dashboardId } = this.state;

    return name !== '' && this.validateRecipients() && dashboardId > 0;
  }

  render() {
    const { heading } = this.props;
    const { message, name, recipients, replyTo, schedule, dashboardId, dashboardName } = this.state;

    const labelWidth = 10;
    const inputWidth = 30;

    const currentDashboard = dashboardId > 0 ? { value: dashboardId, label: dashboardName } : undefined;
    const canPreview = dashboardId > 0;

    return (
      <div>
        <h3 className="page-sub-heading">{heading}</h3>
        <form className="gf-form-group" onSubmit={this.submitForm}>
          <FormField
            label="Name"
            labelWidth={labelWidth}
            inputWidth={inputWidth}
            placeholder="System status report"
            value={name}
            onChange={(event: ChangeEvent<HTMLInputElement>) => this.onChange('name', event.target.value)}
            required={true}
          />
          <div className="gf-form-inline">
            <FormLabel width={labelWidth}>Choose dashboard</FormLabel>
            <DashboardPicker
              onSelected={this.onDashboardChange}
              className="width-30"
              currentDashboardId={currentDashboard}
            />
          </div>
          <div className="form-field">
            <FormLabel width={labelWidth}>Recipients</FormLabel>
            <textarea
              className={`gf-form-input width-${inputWidth}`}
              placeholder="name@company.com;another.name@company.com;"
              rows={1}
              value={recipients}
              onChange={(event: ChangeEvent<HTMLTextAreaElement>) => this.onChange('recipients', event.target.value)}
              onBlur={this.validateRecipients}
              required={true}
            />
          </div>
          <FormField
            label="Reply to"
            labelWidth={labelWidth}
            inputWidth={inputWidth}
            placeholder="your.address@company.com - optional"
            value={replyTo}
            onChange={(event: ChangeEvent<HTMLInputElement>) => this.onChange('replyTo', event.target.value)}
            type="email"
          />
          <div className="form-field">
            <FormLabel width={labelWidth}>Custom message</FormLabel>
            <textarea
              value={message}
              onChange={(event: ChangeEvent<HTMLTextAreaElement>) => this.onChange('message', event.target.value)}
              className={`gf-form-input width-${inputWidth}`}
              rows={10}
            />
          </div>
        </form>
        <div className="gf-form-group">
          <a
            href={`/api/reports/render/pdf/${dashboardId}`}
            className={`btn btn-secondary ${!canPreview ? 'disabled' : ''}`}
            target="_blank"
          >
            Preview
          </a>
        </div>
        <div className="gf-form-group">
          <div>
            <h4>Scheduling</h4>
            <ReportScheduling
              schedulingOptions={schedule}
              onModeChange={this.onModeChange}
              onDayOfWeekChange={this.onDayOfWeekChange}
              onTimeOfDayChange={this.onTimeOfDayChange}
            />
          </div>
        </div>
        <div className="gf-form-button-row">
          <button className={`btn btn-primary ${this.isFormValid() ? '' : 'disabled'}`} onClick={this.submitForm}>
            Save
          </button>
        </div>
      </div>
    );
  }
}
